// Gregory Ecker
// 7/13/94 (revised 7/17 to include perspective projection.)
// Graphics programming
// Prof. Velho

#include "matrix.h"

// member functions defined for matrix + vector classes.
//
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// matrix multiply.
//	returned matrix is the dot product of *this object (on the left
//	of the * operator) and the parameter (on the right side of the
//	* operator).
matrix  matrix::operator*(matrix& m) {
	matrix x;
	float a,b;
	for(int i=0;i<4;i++){
	 for(int j=0;j<4;j++){
	    x(i,j)=0;
	    for(int k=0;k<4;k++){
	       x(i,j) += m(k,j) * mat4x4[i][k];
	      }
	    }
	  }
	  return(x);
};
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Ymirror :
//  	returns mirror matrix to produce mirror images by reflection
//	about y axis.
//
matrix& matrix::Ymirror(void){


  mat4x4[0][2]=1;mat4x4[0][3]=0;mat4x4[0][1]=0;mat4x4[0][0]=0;

  mat4x4[1][0]=0;mat4x4[1][1]=-1;mat4x4[1][2]=0;mat4x4[1][3]=0;

  mat4x4[2][0]=0;mat4x4[2][1]=0;mat4x4[2][2]=-1;mat4x4[2][3]=0;

  mat4x4[3][0]=0;mat4x4[3][1]=0;mat4x4[3][2]=0;mat4x4[3][3]=1;
  return *this;
};
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Identity :  returns identity matrix
matrix& matrix::Identity(void){
// identity matrix looks like :
// 1 0 0 0
// 0 1 0 0
// 0 0 1 0
// 0 0 0 1

  mat4x4[0][0]=1;mat4x4[0][1]=0;mat4x4[0][2]=0;mat4x4[0][3]=0;

  mat4x4[1][0]=0;mat4x4[1][1]=1;mat4x4[1][2]=0;mat4x4[1][3]=0;

  mat4x4[2][0]=0;mat4x4[2][1]=0;mat4x4[2][2]=1;mat4x4[2][3]=0;

  mat4x4[3][0]=0;mat4x4[3][1]=0;mat4x4[3][2]=0;mat4x4[3][3]=1;
  return *this;
};

//////////////////////////////////////////////////////////////////////////
// added ## 7/17
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Project :
//	returns perspective projection matrix ASSUMING that :
//	the center of projection is about the origin.
//	 d is the distance to the center of the projection.

matrix& matrix::Project(float d){
  mat4x4[0][0]=1;             mat4x4[1][0]= 0;
  mat4x4[0][1]=0;             mat4x4[1][1]= 1;
  mat4x4[0][2]=0;             mat4x4[1][2]= 0;
  mat4x4[0][3]=0;             mat4x4[1][3]= 0;

  mat4x4[2][0]=0;             mat4x4[3][0]=0;
  mat4x4[2][1]=0;	      mat4x4[3][1]=0;
  mat4x4[2][2]=0;	      mat4x4[3][2]=0;   // 1/d
  mat4x4[2][3]=1/d;           mat4x4[3][3]=1;
  return *this;
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Xrotate:  calculates a matrix to rotate a vector about the X axis
// counterclockwise.
//      Returns the apropriate matrix and sets *this object to that matrix.
//      This function is dependent on <MATH.H>
matrix& matrix::Xrotate(float theta){
  // The angle theta is measured in radians.
  // the x-rotation matrix is:
  //    1    0      0      0
  //    0  cos(@) -sin(@)  0
  //    0  sin(@)  cos(@)  0
  //    0       0    0     1

  mat4x4[0][0]=1;             mat4x4[1][0]= 0;
  mat4x4[0][1]=0;             mat4x4[1][1]= ::cos(theta);
  mat4x4[0][2]=0;             mat4x4[1][2]= -::sin(theta);
  mat4x4[0][3]=0;             mat4x4[1][3]= 0;

  mat4x4[2][0]=0;             mat4x4[3][0]=0;
  mat4x4[2][1]=::sin(theta);  mat4x4[3][1]=0;
  mat4x4[2][2]=::cos(theta);  mat4x4[3][2]=0;
  mat4x4[2][3]=0;             mat4x4[3][3]=1;

  return(*this);
};
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// NXrotate:  calculates a matrix to rotate a vector about the X axis
// clockwise.
//      Returns the apropriate matrix and sets *this object to that matrix.
//      This function is dependent on <MATH.H>
matrix& matrix::NXrotate(float theta){
  // The angle theta is measured in radians.
  // the Nx-rotation matrix is:
  //    1    0      0      0
  //    0  cos(@)  sin(@)  0
  //    0  -sin(@)  cos(@)  0
  //    0       0    0     1

  mat4x4[0][0]=1;             mat4x4[1][0]= 0;
  mat4x4[0][1]=0;             mat4x4[1][1]= ::cos(theta);
  mat4x4[0][2]=0;             mat4x4[1][2]= ::sin(theta);
  mat4x4[0][3]=0;             mat4x4[1][3]= 0;

  mat4x4[2][0]=0;             mat4x4[3][0]=0;
  mat4x4[2][1]=-::sin(theta); mat4x4[3][1]=0;
  mat4x4[2][2]=::cos(theta);  mat4x4[3][2]=0;
  mat4x4[2][3]=0;             mat4x4[3][3]=1;

  return(*this);
};


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Yrotate:  calculates a matrix to rotate a vector about the Y axis
//	(clockwise).
// 	returns the apropriate matrix and sets *this object to that matrix.
//      This function is dependent on <MATH.H>
matrix& matrix::Yrotate(float theta){
  // The angle theta is measured in radians.
  // the y-rotation matrix is:
  //   cos(@)   0   sin(@)   0
  //     0      1     0      0
  //   -sin(@)  0   cos(@)   0
  //     0      0     0      1

  mat4x4[0][0]=::cos(theta);  mat4x4[1][0]= 0;
  mat4x4[0][1]=0; 	      mat4x4[1][1]= 1;
  mat4x4[0][2]=::sin(theta);  mat4x4[1][2]= 0;
  mat4x4[0][3]=0;             mat4x4[1][3]= 0;

  mat4x4[2][0]=-::sin(theta); mat4x4[3][0]=0;
  mat4x4[2][1]=0;             mat4x4[3][1]=0;
  mat4x4[2][2]=::cos(theta);  mat4x4[3][2]=0;
  mat4x4[2][3]=0;             mat4x4[3][3]=1;

  return(*this);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// NYrotate:  calculates a matrix to rotate a vector about the Y axis
//	(counterclockwise).
// 	returns the apropriate matrix and sets *this object to that matrix.
//      This function is dependent on <MATH.H>
matrix& matrix::NYrotate(float theta){
  // The angle theta is measured in radians.
  // the y-rotation matrix is:
  //   cos(@)   0   -sin(@)  0
  //     0      1     0      0
  //   sin(@)   0   cos(@)   0
  //     0      0     0      1

  mat4x4[0][0]=::cos(theta);  mat4x4[1][0]= 0;
  mat4x4[0][1]=0; 	      mat4x4[1][1]= 1;
  mat4x4[0][2]=-::sin(theta);  mat4x4[1][2]= 0;
  mat4x4[0][3]=0;             mat4x4[1][3]= 0;

  mat4x4[2][0]=::sin(theta);  mat4x4[3][0]=0;
  mat4x4[2][1]=0;             mat4x4[3][1]=0;
  mat4x4[2][2]=::cos(theta);  mat4x4[3][2]=0;
  mat4x4[2][3]=0;             mat4x4[3][3]=1;

  return(*this);
};


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Zrotate:  calculates a matrix to rotate a vector about the Z axis
//	(clockwise).
// 	returns the apropriate matrix and sets *this object to that matrix.
//      This function is dependent on <MATH.H>
matrix& matrix::Zrotate(float theta){
  // The angle theta is measured in radians.
  // the z-rotation matrix is:
  //   cos(@) -sin(@) 0   0
  //   sin(@)  cos(@) 0   0
  //     0       0    1   0
  //     0       0    0   1

  mat4x4[0][0]=::cos(theta);  mat4x4[1][0]= ::sin(theta);
  mat4x4[0][1]=-::sin(theta); mat4x4[1][1]= ::cos(theta);
  mat4x4[0][2]=0;             mat4x4[1][2]= 0;
  mat4x4[0][3]=0;             mat4x4[1][3]= 0;

  mat4x4[2][0]=0;            mat4x4[3][0]=0;
  mat4x4[2][1]=0;            mat4x4[3][1]=0;
  mat4x4[2][2]=1;            mat4x4[3][2]=0;
  mat4x4[2][3]=0;            mat4x4[3][3]=1;

  return(*this);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// NZrotate:  calculates a matrix to rotate a vector about the Z axis
//	(counterclockwise).
// 	returns the apropriate matrix and sets *this object to that matrix.
//      This function is dependent on <MATH.H>
matrix& matrix::NZrotate(float theta){
  // The angle theta is measured in radians.
  // the z-rotation matrix is:
  //   cos(@)  sin(@) 0   0
  //   -sin(@) cos(@) 0 `  0
  //     0       0    1   0
  //     0       0    0   1

  mat4x4[0][0]=::cos(theta);  mat4x4[1][0]= -::sin(theta);
  mat4x4[0][1]=::sin(theta);  mat4x4[1][1]= ::cos(theta);
  mat4x4[0][2]=0;             mat4x4[1][2]= 0;
  mat4x4[0][3]=0;             mat4x4[1][3]= 0;

  mat4x4[2][0]=0;            mat4x4[3][0]=0;
  mat4x4[2][1]=0;            mat4x4[3][1]=0;
  mat4x4[2][2]=1;            mat4x4[3][2]=0;
  mat4x4[2][3]=0;            mat4x4[3][3]=1;

  return(*this);
};


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// matrix equality operator
//	returns 1 if two matrices are equal, 0 otherwise.
int matrix::operator==(const matrix& m) const{
      int equal=TRUE,i=0,j=0,count=0;
      while((equal==TRUE)&&(i<4)){
	while((equal==TRUE)&&(j<4)){
	    if(mat4x4[i][j]!=m(i,j)){
	      equal=FALSE;
	     }
	    count++;
	    j++;
	   }
	   i++;
	 }
     return (equal);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// matrix inequality operator
//	returns 0 if two matrices are equal, 1 otherwise.
int matrix::operator!=(const matrix& m) const{
      int equal=TRUE,i=0,j=0,count=0;
      while((equal==TRUE)&&(i<4)){
	while((equal==TRUE)&&(j<4)){
	    if(mat4x4[i][j]!=m(i,j)){
	      equal=FALSE;
	     }
	    count++;
	    j++;
	   }
	   i++;
	 }
     return (!equal);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// matrix assignment operator.
//	copies one matrix into *this one.
void matrix::operator=(const matrix& m){      // vector assignment op.
     for(int i=0;i<4;i++){
       for(int j=0;j<4;j++){
	  mat4x4[i][j]=m(i,j);
	}
      }
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// scale:    returns scaling matrix according to amounts specified
//	by parameters.  Also changes *this object to specified scaling
//	matrix.  parameters dy,dz default to 1 (no scaling) if their
//	values are not given.

matrix& matrix::scale(float sx,float sy=1,float sz=1){
 // scaling matrix should look like :
 //
 //     sx 0  0  0
 //	0 sy  0  0
 //     0  0 sz  0
 //     0  0  0  1
 mat4x4[0][0]= sx; mat4x4[1][0]=0;  mat4x4[2][0]=0;  mat4x4[3][0]=0;
 mat4x4[0][1]= 0;  mat4x4[1][1]=sy; mat4x4[2][1]=0;  mat4x4[3][1]=0;
 mat4x4[0][2]= 0;  mat4x4[1][2]=0;  mat4x4[2][2]=sz; mat4x4[3][2]=0;
 mat4x4[0][3]= 0;  mat4x4[1][3]=0;  mat4x4[2][3]=0;  mat4x4[3][3]=1;
 return(*this);
};


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// scale:    returns scaling matrix according to amounts specified
//	by vector parameter.  Also changes *this object to specified scaling
//	matrix.

matrix& matrix::scale(vector v){
 // scaling matrix should look like :
 //
 //     sx 0  0  0
 //	0 sy  0  0
 //     0  0 sz  0
 //     0  0  0  1
 mat4x4[0][0]= v[0]; mat4x4[1][0]=0;  mat4x4[2][0]=0;  mat4x4[3][0]=0;
 mat4x4[0][1]= 0;  mat4x4[1][1]=v[1]; mat4x4[2][1]=0;  mat4x4[3][1]=0;
 mat4x4[0][2]= 0;  mat4x4[1][2]=0;  mat4x4[2][2]=v[2]; mat4x4[3][2]=0;
 mat4x4[0][3]= 0;  mat4x4[1][3]=0;  mat4x4[2][3]=0;  mat4x4[3][3]=1;
 return(*this);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// translate::   returns translation matrix according to amounts specified
//     by x,y,z parameters (expressed as difference from current position.)
//    also changes *this object to the apropriate translation matrix.
//    dz may be left out and is thus assumed to be 0.
//    if 2 param.s are left out, they are assumed to be dy=0,dz=0

matrix& matrix::translate(float dx,float dy=0,float dz=0){
   // translation matrix should loog like :
   //      1 0 0 0    dx  -- amount to change in x direction.
   //      0 1 0 0    dy  -- amount to change in y direction.
   //      0 0 1 0    dz  -- amount to change in z direction.
   //    dx dy dz 1
   //
 mat4x4[0][0]=1;  mat4x4[1][0]=0;  mat4x4[2][0]=0;  mat4x4[3][0]=dx;
 mat4x4[0][1]=0;  mat4x4[1][1]=1;  mat4x4[2][1]=0;  mat4x4[3][1]=dy;
 mat4x4[0][2]=0;  mat4x4[1][2]=0;  mat4x4[2][2]=1;  mat4x4[3][2]=dz;
 mat4x4[0][3]=0;  mat4x4[1][3]=0;  mat4x4[2][3]=0;  mat4x4[3][3]=1;
 return (*this);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// translate::   returns translation matrix according to amounts specified
//     by the vector parameter.
//    also changes *this object to the apropriate translation matrix.

matrix& matrix::translate(vector v){
   // translation matrix should loog like :
   //      1 0 0 dx  -- amount to change in x direction.
   //      0 1 0 dy  -- amount to change in y direction.
   //      0 0 1 dz  -- amount to change in z direction.
   //      0 0 0 1
   //
 mat4x4[0][0]=1;  mat4x4[1][0]=0;  mat4x4[2][0]=0;  mat4x4[3][0]=0;
 mat4x4[0][1]=0;  mat4x4[1][1]=1;  mat4x4[2][1]=0;  mat4x4[3][0]=0;
 mat4x4[0][2]=0;  mat4x4[1][2]=0;  mat4x4[2][2]=1;  mat4x4[3][0]=0;
 mat4x4[0][3]=v[0]; mat4x4[1][3]=v[1]; mat4x4[2][3]=v[2]; mat4x4[3][1]=1;
 return (*this);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// operator ()  -- same as 2D array dereference ' [][] '
//	used to avoid conflicts in operator overloads.
float& matrix::operator()(int row,int col){
       return(mat4x4[row][col]);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// operator ()  -- same as 2D array dereference ' [][] '
//	used to avoid conflicts in operator overloads.
//  returns lval (for constant objects);
float matrix::operator()(int row,int col) const {
       return(mat4x4[row][col]);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// default matrix constructor:
//     initialize all elements to 0.

matrix::matrix(){
       for(int i=0;i<4;i++){
	  for(int j=0;j<4;j++){
	     mat4x4[i][j]=0.0;
	   }
	 }
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// alternate matrix constructor :
//	takes a 4 by 4 array of floats to be used in the initialization.

matrix::matrix(const float array[4][4]){
	for(int i=0;i<4;i++){
	   for(int j=0;i<4;i++){
	      mat4x4[i][j]=array[i][j];
	     }
	   }
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// matrix copy constructor:
//	initializes new matrix to equal existing one.

matrix::matrix(matrix& m){
	for(int i=0;i<4;i++){
	   for(int j=0;i<4;i++){
	      mat4x4[i][j]=m(i,j);
	     }
	   }
};
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// vector X vector multiply
// resultant scalar is the dot product of *this vector with external vector
//  (primarily to measure distances.

float  vector::operator*(vector& v)const {
	 float retval=0.0;
	 retval=coords[0]*v[0];
	 retval+=coords[1]*v[1];
	 retval+=coords[2]*v[2];
	 retval+=coords[3]*v[3];
       return retval;
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// vector X matrix multiply:
// resultant vector is the dot product of the row vector and matrix parameter.

vector vector::operator*(matrix& m)const{   // ## return reference to vector ?
	vector v(0,0,0,1);
	// (x0,x1,x2,x3) *  | a b c d |
	//                  | e f g h |
	//                  | i j k l |
	//                  | m n o p |

	// x0 = ax0 + ex1 + ix2 + mx3
	v[0] = ((m(0,0) * coords[0]) + (m(1,0) * coords[1])
	     + (m(2,0) * coords[2]) + (m(3,0) * coords[3]));
	// x1 = bx0 + fx1 + jx2 + nx3
	v[1] = ((m(0,1) * coords[0]) + (m(1,1) * coords[1])
	     + (m(2,1) * coords[2]) + (m(1,3) * coords[3]));
	// x2 = cx0 + gx1 + kx2 + ox3
	v[2] = ((m(0,2) * coords[0]) + (m(1,2) * coords[1])
	     + (m(2,2) * coords[2]) + (m(3,2) * coords[3]));
	// x3 = dx0 + hx1 + lx2 + px3
	v[3] = ((m(0,3) * coords[0]) + (m(1,3) * coords[1])
	     + (m(2,3) * coords[2]) + (m(3,3) * coords[3]));
	return (v);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// copy constructor:  instantiates new vector with same coord.s as param.

vector::vector(const vector& v){
      for(int i=0;i<4;i++){
       coords[i]=v.coords[i];
      }
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// default constructor :
//	sets all coordinates to 0, magnitude to 1

vector::vector(){
	for(int i=0;i<3;i++){
		coords[i]=0.0;
	   };
	coords[3]=1.0;
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// alternate constructor :
//	sets all coordinates, magnitude to parameters.

vector::vector(float x,float y,float z,float mag){
	coords[0]=x;
	coords[1]=y;
	coords[2]=z;
	coords[3]=mag;
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// vector de-reference : allows assignment of individual scalars

float& vector::operator[](const int i){
	return( coords[i]);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// vector de-reference : returns specified scalar (to be used as Rval only).

float vector::operator[](const int i) const{
	return( coords[i]);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// vector compare : assumes normalized vectors.
//		    returns 1 if coord.s are different, 0 otherwise.

int vector::operator!=(const vector& v) const{
	 int i=0,equal=TRUE;
	 while((equal==TRUE)&&(i<4)){
		if(coords[i]!=v[i]){
		  equal=FALSE;
		 }
		i++;
	     };
	 return (!equal);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// vector compare : assumes normalized vectors.
//		    returns 1 if coord.s are the same, 0 otherwise.

int vector::operator==(const vector& v) const{
	 int i=0,equal=TRUE;
	 while((equal==TRUE)&&(i<4)){
		if(coords[i]!=v[i]){
		  equal=FALSE;
		 }
		i++;
	     };
	 return (equal);
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// vector assignment : sets vector object to same coord.s as parameter.

vector& vector::operator=(const vector& v){
	coords[0]=v[0];
	coords[1]=v[1];
	coords[2]=v[2];
	coords[3]=v[3];
     return *this;
};